package riven.virtualthreads;

import java.util.concurrent.atomic.AtomicLong;

import de.matthiasmann.continuations.Coroutine;
import de.matthiasmann.continuations.CoroutineProto;
import de.matthiasmann.continuations.SuspendExecution;
import de.matthiasmann.continuations.Coroutine.State;

/*
 * Created on Nov 30, 2010
 * 
 * @author Riven
 */

public class VirtualThread implements CoroutineProto, VirtualRunnable {
   
   private static final AtomicLong ID_PROVIDER = new AtomicLong();
   
   private final VirtualRunnable   task;
   private final Coroutine         coroutine;
   private final long              id;
   
   public VirtualThread() {
      this(null);
   }
   
   public VirtualThread(VirtualRunnable task) {
      this.task = task;
      this.coroutine = new Coroutine(this);
      this.id = ID_PROVIDER.incrementAndGet();
      this.state = VirtualThreadState.NEW;
   }
   
   //
   
   private VirtualProcessor proc;
   
   public VirtualProcessor getProcessor() {
      return assertNotNull(this.proc);
   }
   
   //
   
   public void start() {
      this.start(VirtualProcessor.LAST_DEFINED.get());
   }
   
   public void start(VirtualProcessor proc) {
      assertNotNull(proc);
      assertNull(this.proc);
      assertTrue(this.state == VirtualThreadState.NEW);
      
      this.proc = proc;
      this.state = VirtualThreadState.RUNNABLE;
      this.wakeUpAt = 0L;
      this.proc.enqueue(this);
   }
   
   public void spawn(VirtualRunnable task) {
      new VirtualThread(task).start(this.proc);
   }
   
   public void resume() {
      if (this.state == VirtualThreadState.SUSPENDED) {
         this.state = VirtualThreadState.RUNNABLE;
         this.wakeUpAt = 0L;
         this.proc.enqueue(this);
      }
   }
   
   public void kill() throws SuspendExecution {
      if (this.state == VirtualThreadState.SUSPENDED) {
         this.state = VirtualThreadState.TERMINATED;
      }
      else if (this.state == VirtualThreadState.RUNNABLE) {
         this.state = VirtualThreadState.TERMINATED;
         if (!this.getProcessor().unschedule(this)) {
            throw new IllegalStateException("thread not scheduled");
         }
      }
   }
   
   private VirtualThreadState state;
   long                       wakeUpAt;
   
   @Override
   public final void coExecute() throws SuspendExecution {
      this.run();
   }
   
   public void run() throws SuspendExecution {
      this.task.run();
   }
   
   VirtualThreadState step() {
      if (this.state != VirtualThreadState.RUNNABLE) {
         return this.state;
      }
      
      coroutine.run();
      
      if (coroutine.getState() == State.FINISHED) {
         this.state = VirtualThreadState.TERMINATED;
      }
      
      return this.state;
   }
   
   public long getId() {
      return this.id;
   }
   
   @Override
   public int hashCode() {
      return (int) this.id;
   }
   
   //
   
   public static void yield() throws SuspendExecution {
      sleepImpl(0L);
   }
   
   public static void sleep(long ms) throws SuspendExecution {
      sleepImpl(Math.max(ms, 0L));
   }
   
   public static void wakeup(long at) throws SuspendExecution {
      wakeupImpl(at);
   }
   
   public static void suspend() throws SuspendExecution {
      stateImpl(VirtualThreadState.SUSPENDED);
   }
   
   public static void stop() throws SuspendExecution {
      stateImpl(VirtualThreadState.TERMINATED);
   }
   
   private static void sleepImpl(long ms) throws SuspendExecution {
      VirtualThread thread = currentThread();
      thread.wakeUpAt = thread.getProcessor().getCurrentTime() + ms;
      Coroutine.yield();
   }
   
   private static void wakeupImpl(long at) throws SuspendExecution {
      currentThread().wakeUpAt = at;
      Coroutine.yield();
   }
   
   private static void stateImpl(VirtualThreadState state) throws SuspendExecution {
      currentThread().state = state;
      Coroutine.yield();
   }
   
   public static final VirtualThread currentThread() {
      return assertNotNull((VirtualThread) Coroutine.getActiveCoroutine().getProto());
   }
   
   //
   
   private static <T> T assertNotNull(T value) {
      if (value == null) {
         throw new IllegalArgumentException("argument must not be null");
      }
      return value;
   }
   
   private static <T> T assertNull(T value) {
      if (value != null) {
         throw new IllegalArgumentException("argument must be null");
      }
      return value;
   }
   
   private static boolean assertTrue(boolean value) {
      if (!value) {
         throw new IllegalArgumentException("argument must be true");
      }
      return value;
   }
}
